from typing import Any, Optional, TypedDict
import requests

class HttpApiConfig(TypedDict):
    agentlabs_url: str;
    project_id: str;
    secret: str;

class HttpApi:
    def __init__(self, agentlabs_url: str, project_id: str, secret: str) -> None:
        self.agentlabs_url = agentlabs_url
        self.project_id = project_id
        self.secret = secret
        self.client = requests.Session()
        self.client.params = {
            "projectId": self.project_id
        }
        self.client.headers.update({
            "x-agentlabs-sdk-secret": self.secret,
            "x-agentlabs-project-id": self.project_id,
        })


    def upload(self, path: str, filename: str, data: bytes, mime_type: Optional[str]) -> Any:
        mime_type = mime_type or "application/octet-stream"
        response = self.client.post(self._make_url(path), files={
            "file": (filename, data, mime_type)
        })
        response.raise_for_status()

        return response.json()

    def download(self, path: str) -> bytes:
        response = self.client.get(self._make_url(path))
        response.raise_for_status()

        return response.content

    def _make_url(self, path: str):
        return f"{self.agentlabs_url}/api{path}"



from datetime import datetime
import os
from typing import Any, Callable
import socketio

class RealtimeClient:
    agent_namespace = '/agent'
    is_connected: bool = False

    def __init__(self, project_id: str, secret: str, url: str):
        self._project_id = project_id
        self._secret = secret
        self._url = url
        self._is_debug_enabled = bool(os.environ.get('DEBUG', False))
        self._io = socketio.Client(logger=self._is_debug_enabled, engineio_logger=self._is_debug_enabled)
    
    def connect(self):
        try:
            self._io.connect(self._url, 
                headers={
                    'x-agentlabs-project-id': self._project_id,
                    'x-agentlabs-sdk-secret': self._secret
                },
                namespaces=[self.agent_namespace],
                transports=['websocket']
            )
            self.is_connected = True
        except Exception as e:
            print(e)
            raise Exception('Failed to connect backend to AgentLabs')

    def disconnect(self):
        if not self.is_connected:
            raise Exception('Not connected')
        self._io.disconnect()

    def on(self, event: str, callback: Callable[[Any], Any]):
        self._io.on(event, callback, namespace=self.agent_namespace)

    def emit(self, event: str, data: Any):
        ts = datetime.now().isoformat()
        self._io.emit(event, {
            "timestamp": ts,
            "data": data
        }, namespace=self.agent_namespace)

    def wait(self):
        self._io.wait()



from enum import Enum
from typing import TypedDict

from agentlabs._internals.http import HttpApi
from .member import _ApiMember, Member

class _MessageAttachment(TypedDict):
    id: str;
    createdAt: str;
    name: str;
    mimeType: str;
    checksumSha256: str;
    driver: str;
    sizeBytes: str;

class _ChatMessage(TypedDict):
    text: str
    conversationId: str
    messageId: str
    agentId: str
    member: _ApiMember
    attachments: list[_MessageAttachment]

class MessageFormat(Enum):
    PLAIN_TEXT = "PLAIN_TEXT"
    MARKDOWN = "MARKDOWN"

class IncomingChatMessage:
    """Represents a chat message received from the AgentLabs server.
    """
    def __init__(self, _http: HttpApi, message: _ChatMessage):
        self.text = message["text"]
        self.conversation_id = message["conversationId"]
        self.message_id = message["messageId"]
        self.member_id = message["member"]["id"]
        self.sender = Member(message["member"])
        self.attachments = [MessageAttachment(_http, attachment) for attachment in message["attachments"]]


class MessageAttachment:
    """Represents a file attachment sent in a chat message.
    """
    def __init__(self, _http: HttpApi, attachment: _MessageAttachment):
        self._http = _http
        self.id = attachment["id"]
        self.created_at = attachment["createdAt"]
        self.name = attachment["name"]
        self.mime_type = attachment["mimeType"]
        self.checksum_sha256 = attachment["checksumSha256"]
        self.size_bytes = attachment["sizeBytes"]

    def download(self) -> bytes:
        return self._http.download(f"/attachments/downloadById/{self.id}")

    def download_to_file(self, file_path: str) -> None:
        data = self.download()
        with open(file_path, "wb") as file:
            file.write(data)



import os
from typing import TypedDict, Optional

MAX_ATTACHMENT_SIZE = 1024 * 1024 * 10

class AttachmentItem:
    def __init__(self):
        pass

class _UploadedAttachment(TypedDict):
    id: str

class Attachment:
    @staticmethod
    def from_local_file(path: str, mime_type: Optional[str] = None) -> AttachmentItem:
        return _LocalFileAttachment(path, mime_type)

    @staticmethod
    def from_buffer(data: bytes, filename: str, mime_type: Optional[str] = None) -> AttachmentItem:
        return _BufferFileAttachment(data, filename, mime_type)

class _LocalFileAttachment(AttachmentItem):
    data: bytes

    def __init__(self, path: str, mime_type: Optional[str]):
        self.path = path
        self.filename = os.path.basename(path)
        self.mime_type = mime_type

    def load(self):
        file = open(self.path, 'rb')
        data = file.read()

        if len(data) > MAX_ATTACHMENT_SIZE:
            raise Exception("Attachment too big, max size is 10MB")

        self.data = data

class _BufferFileAttachment(AttachmentItem):
    data: bytes

    def __init__(self, data: bytes, filename: str, mime_type: Optional[str]):
        self.filename = filename
        self.data = data
        self.mime_type = mime_type
    
    def load(self):
        if len(self.data) > MAX_ATTACHMENT_SIZE:
            raise Exception("Attachment too big, max size is 10MB")

        pass



from typing import Set
from uuid import uuid4

from agentlabs._internals.http import HttpApi

from .attachment import AttachmentItem, _LocalFileAttachment, _UploadedAttachment, _BufferFileAttachment
from .chat import MessageFormat
from ._internals.realtime import RealtimeClient
from ._internals.utils import chunk_str
from ._internals.const import (
    DEFAULT_STREAM_TOKEN_SIZE,
    DEFAULT_MESSAGE_TYPING_INTERVAL_MS,
    DEFAULT_INITIAL_MESSAGE_LOADING_DELAY_MS
)
import time

class AgentStream:
    """Creates a stream to send a message in multiple parts.
        This is done by making successive calls to write() and then calling end() when done.
    """
    is_ended: bool = False

    def __init__(self, realtime: RealtimeClient, agent_id: str, conversation_id: str, format: MessageFormat):
        self.conversation_id = conversation_id
        self.message_id = str(uuid4())
        self.format = format
        self._realtime = realtime
        self.agent_id = agent_id
        self._start(self.message_id)

    def _start(self, token: str):
        if self.is_ended:
            raise Exception("Cannot write to a stream that has already been ended.")

        self._realtime.emit('stream-chat-message-start', {
            "conversationId": self.conversation_id,
            "messageId": self.message_id,
            "attachments": [],
            "format": self.format.value,
            "agentId": self.agent_id
        })

    def write(self, token: str):
        """Writes a token to the stream. This can be used to send a message in multiple parts.
        Writing to a stream on which end() has been called will raise an exception.
        """
        if self.is_ended:
            raise Exception("Cannot write to a stream that has already been ended.")

        self._realtime.emit('stream-chat-message-token', {
            "conversationId": self.conversation_id,
            "messageId": self.message_id,
            "text": token,
            "attachments": [],
            "format": self.format.value,
            "agentId": self.agent_id
        })


    def typewrite(
        self,
        token: str,
        interval_ms: int = DEFAULT_MESSAGE_TYPING_INTERVAL_MS
        ):
        """Sends a stream to a conversation, with a typewriter animation.
        Such message will be displayed smoothly to the user with a nice typewriter animation.
        """
        chunks = chunk_str(token, DEFAULT_STREAM_TOKEN_SIZE)
        for chunk in chunks:
            self.write(chunk)
            time.sleep(interval_ms / 1000)

        self.end()

    def end(self):
        """
        Ends the stream. After a stream is ended it cannot be written to anymore, and
        doing so will raise an exception.
        """
        self.is_ended = True
        self._realtime.emit('stream-chat-message-end', {
            "conversationId": self.conversation_id,
            "messageId": self.message_id,
            "agentId": self.agent_id
        })

class Agent:
    def __init__(self, id: str, realtime: RealtimeClient, http: HttpApi) -> None:
        self._realtime = realtime
        self._http = http
        self.id = id

    def _upload_attachment(self, attachment: AttachmentItem) -> _UploadedAttachment:
        uploaded_attachment = None

        if isinstance(attachment, _LocalFileAttachment) or isinstance(attachment, _BufferFileAttachment):
            attachment.load()
            uploaded_attachment = self._http.upload('/attachments/uploadSync',
                    data=attachment.data,
                    filename=attachment.filename,
                    mime_type=attachment.mime_type
            )

        if uploaded_attachment is None:
            raise Exception("Could not upload attachment")

        return uploaded_attachment

    def echart(self, conversation_id: str, echart: dict, text = '', text_format: MessageFormat = MessageFormat.PLAIN_TEXT):
        """Sends an [echart](https://echarts.apache.org/en/index.html) chart to a conversation, with an optional text message attached.

        The echart parameter holds the chart configuration, including styling and data, and should only contain JSON serializable values.
        Because echart is a Javascript library, this python SDK does not provide strong typing for the echart parameter.

        You may want to refer to the examples from [this page](https://echarts.apache.org/examples/en/index.html) to get an idea of what the echart parameter should look like.
        """
        self._realtime.emit('chat-message', {
            "conversationId": conversation_id,
            "text": text,
            "agentId": self.id,
            "source": "AGENT",
            "format": text_format.value,
            "attachments": [],
            "type": "ECHART",
            "metadata": echart
        });

    def send(self, text: str, conversation_id: str, format: MessageFormat = MessageFormat.PLAIN_TEXT, attachments: list[AttachmentItem] = []):
        """Sends a message to a conversation.
        Such message will be communicated instantly to the conversation.
        """
        uploaded_attachments: list[_UploadedAttachment] = [self._upload_attachment(attachment) for attachment in attachments]

        self._realtime.emit('chat-message', {
            "conversationId": conversation_id,
            "text": text,
            "agentId": self.id,
            "source": "AGENT",
            "format": format.value,
            "attachments": uploaded_attachments
        })

    def create_stream(self, conversation_id: str, format: MessageFormat) -> AgentStream:
        """Creates a stream to send a message in multiple parts.
        This is done by making successive calls to write() and then calling end() when done.
        """
        return AgentStream(self._realtime, self.id, conversation_id, format)

    def typewrite(
        self,
        text: str,
        conversation_id: str,
        format: MessageFormat = MessageFormat.PLAIN_TEXT,
        initial_delay_ms: int = DEFAULT_INITIAL_MESSAGE_LOADING_DELAY_MS,
        interval_ms: int = DEFAULT_MESSAGE_TYPING_INTERVAL_MS
        ):
        """Sends a message to a conversation, with a typewriter animation.
        Such message will be displayed smoothly to the user with a nice typewriter animation and loading
        indicator.
        """
        stream = self.create_stream(conversation_id, format)

        if (initial_delay_ms > 0):
            time.sleep(initial_delay_ms / 1000)

        stream.typewrite(text, interval_ms)

        stream.end()



from typing import Optional, TypedDict


class _ApiMember(TypedDict):
    id: str
    createdAt: str
    updatedAt: str

    verifiedAt: Optional[str]
    bannedAt: Optional[str]

    firstName: Optional[str]
    lastName: Optional[str]
    fullName: Optional[str]

    email: str
    isAnonymous: bool

    profilePictureUrl: Optional[str]
    projectId: str

class Member: 
    def __init__(self, api_member: _ApiMember):
        self.id = api_member["id"]
        self.created_at = api_member["createdAt"]
        self.updated_at = api_member["updatedAt"]
        self.verified_at = api_member["verifiedAt"]
        self.banned_at = api_member["bannedAt"]
        self.first_name = api_member["firstName"]
        self.last_name = api_member["lastName"]
        self.full_name = api_member["fullName"]
        self.email = api_member["email"]
        self.is_anonymous = api_member["isAnonymous"]
        self.profile_picture_url = api_member["profilePictureUrl"]
        self.project_id = api_member["projectId"]
